from enum import IntEnum, auto
from typing import Dict, Tuple


class ModelState(IntEnum):
    """
    Contains the state of each model within a Session scope.

    - UNBOUND indicates that the model is not bound to any Session.
    - CLEAN indicates that the model has been retrieved from the remote server and has
    not been modified. CLEAN models are ignored during a `commit`.
    - NEW indicates that the model has been created within the session scope, but
    not yet exists on the remote server. During `commit`, the model will be passed by
    the session to the DAO's `add` method.
    - DIRTY indicates that the model has been retrieved from the remote server but has
    been modified within the Session scope. During `commit`, the model will be passed
    by the session to the DAO's `update` method.
    - DELETED indicates that the model has been retrieved from the remote server but
    has been deleted within the Session scope. During `commit`, the model will be
    passed by the session to the DAO's `remove` method.
    - DISCARDED indicates that the model has been discarded from the internal Session
    cache and is ready to be cleared up.

    The state management of models is done entirely by the Session instance and stored
    as part of the model instance.
    """

    UNBOUND = auto()
    CLEAN = auto()
    NEW = auto()
    DIRTY = auto()
    DELETED = auto()
    DISCARDED = auto()


class Transition(IntEnum):
    """
    Contains the possible transitions of an object within the Session scope. It is used
    by the ModelStateMachine and Session to figure out the next state of an object.
    """

    REGISTER_OBJECT = auto()
    GET_OBJECT = auto()
    PERSIST_OBJECT = auto()
    ADD_OBJECT = auto()
    UPDATE_OBJECT = auto()
    REMOVE_OBJECT = auto()
    ROLLBACK_OBJECT = auto()
    CLEAN_OBJECT = auto()


class ModelStateMachine:
    """
    Statically defines all possible resulting states of a model with current state
    ModelState going through a transition Transition.
    """

    _transitions: Dict[Tuple[Transition, ModelState], ModelState] = {
        (Transition.REGISTER_OBJECT, ModelState.UNBOUND): ModelState.CLEAN,
        (Transition.GET_OBJECT, ModelState.UNBOUND): ModelState.CLEAN,
        (Transition.ADD_OBJECT, ModelState.UNBOUND): ModelState.NEW,
        (Transition.UPDATE_OBJECT, ModelState.CLEAN): ModelState.DIRTY,
        (Transition.UPDATE_OBJECT, ModelState.DIRTY): ModelState.DIRTY,
        (Transition.CLEAN_OBJECT, ModelState.DIRTY): ModelState.CLEAN,
        (Transition.PERSIST_OBJECT, ModelState.NEW): ModelState.CLEAN,
        (Transition.PERSIST_OBJECT, ModelState.DIRTY): ModelState.CLEAN,
        (Transition.PERSIST_OBJECT, ModelState.DELETED): ModelState.DISCARDED,
        (Transition.REMOVE_OBJECT, ModelState.CLEAN): ModelState.DELETED,
        (Transition.REMOVE_OBJECT, ModelState.DIRTY): ModelState.DELETED,
        (Transition.REMOVE_OBJECT, ModelState.NEW): ModelState.DISCARDED,
        (Transition.ROLLBACK_OBJECT, ModelState.DIRTY): ModelState.CLEAN,
        (Transition.ROLLBACK_OBJECT, ModelState.NEW): ModelState.DISCARDED,
        (Transition.ROLLBACK_OBJECT, ModelState.DELETED): ModelState.DISCARDED,
    }

    @classmethod
    def transition(
        cls, transition: Transition, current_state: ModelState
    ) -> ModelState:
        """
        Maps the Transition `transition` from `current_state` to a new ModelState.

        :param transition: The Transition being made by the Session.
        :param current_state: The current ModelState of the model.
        :return: The resulting model state after the transition.
        """
        key = (transition, current_state)
        return cls._transitions.get(key, current_state)



import setuptools

import restio

with open("README.md", "r") as fh:
    long_description = fh.read()

git_url = "https://github.com/eduardostarling/restio.git"

setuptools.setup(
    name=restio.__name__,
    version=restio.__version__,
    author=restio.__author__,
    author_email=restio.__email__,
    description="restio Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["restio", "restio.fields"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    url=git_url,
    project_urls={
        "Source Code": git_url,
        "Documentation": "https://restio.readthedocs.io/en/latest/",
    },
    python_requires=">=3.7",
)



import types
import weakref
from typing import Any, Callable, Dict, Optional, Set, Union

ListeningMethod = Union[weakref.WeakMethod, types.FunctionType]
CallbackMethod = Union[types.MethodType, types.FunctionType, Callable[..., Any]]


class EventListener:
    """
    Event listener class.

    The events registered in EventListener objects are identified by a string hash
    `event` using the method `subscribe`. The callback functions provided can be normal
    functions or instance methods. Coroutine functions are not supported.

    Callbacks are synchronously triggered by `dispatch` when they match the `event`
    hash. The order in which callbacks are triggered is not deterministic. Extra
    arguments passed to args and kwargs are then propagated to the stored callbacks.

    Methods are stored internally with the wrapper WeakMethod in order to allow for
    garbage collection of instances that are no longer in use. Methods that have been
    garbage collected are ignored if eventually dispatched.
    """

    _listener: Dict[str, Set[ListeningMethod]]

    def __init__(self):
        self._listener = {}

    def subscribe(self, event: str, method: CallbackMethod):
        """
        Subscribes the callback `method` to the event `event`.

        :param event: The event hash.
        :param method: The callback method.
        :raises ValueError: If no event name is provided.
        """
        if not event:
            raise ValueError("You must specify a valid event name.")

        weak_method = self._reference_method(method)
        self._listener.setdefault(event, set())
        self._listener[event].add(weak_method)

    def unsubscribe(self, event: str, method: CallbackMethod):
        """
        Unsubscribes the callback `method` from the event `event`.

        :param event: The event hash.
        :param method: The callback method.
        """
        weak_method = self._reference_method(method)
        if event in self._listener and weak_method in self._listener[event]:
            self._listener[event].remove(weak_method)

    def _reference_method(self, method: CallbackMethod) -> ListeningMethod:
        if isinstance(method, types.MethodType):
            return weakref.WeakMethod(method)
        elif isinstance(method, types.FunctionType):
            return method
        else:
            raise TypeError(
                "The parameter `method` must be either a method or a function"
            )

    def dispatch(self, event: str, *args, **kwargs):
        """
        Synchronously dispatches all callbacks stored with a particular `event` hash.
        `args` and `kwargs` are optional and are passed to the callbacks, therefore
        they should match the signature of all callbacks stored for the event. Errors
        should be handled by the callbacks properly, otherwise raised exceptions will
        interrupt the dispatching and will be propagated back to the caller.

        :param event: The event hash.
        :param args: The positional arguments to be passed into the subscribed methods.
        :param kwargs: The keyword arguments to be passed into the subscribed methods.
        """
        if event in self._listener:
            for weak_method in self._listener[event]:
                method = self._resolve_reference(weak_method)
                if not method:
                    continue

                method(*args, **kwargs)

    def _resolve_reference(
        self, reference: ListeningMethod
    ) -> Optional[CallbackMethod]:
        if isinstance(reference, weakref.WeakMethod):
            return reference()
        else:
            return reference
